/*
 *  cusbandroid.cpp
 *
 */

#include "stdafx.h"
#include "cusb.h"
#include "Java/com_phidgets_usb_Manager.h"
#include "Java/com_phidgets_usb_Phidget.h"

int AndroidUsbJarAvailable = PFALSE;
int AndroidUsbInitialized = PFALSE;

int CUSBCloseHandle(CPhidgetHandle phid)
{
	int ret = 0;
	int result = EPHIDGET_OK;
	
	if(!ANDROID_USB_GOOD)
		return EPHIDGET_UNSUPPORTED;
	
	if (!phid)
		return EPHIDGET_INVALIDARG;
	
	CPhidget_clearStatusFlag(&phid->status, PHIDGET_ATTACHED_FLAG, &phid->lock);
	
	if (phid->deviceHandle == NULL)
		return EPHIDGET_NOTATTACHED;
	
	CThread_join(&phid->readThread);
	
	return CPhidgetAndroid_closePhidget(phid);
}

int CUSBSendPacket(CPhidgetHandle phid, unsigned char *buffer)
{
	if(!ANDROID_USB_GOOD)
		return EPHIDGET_UNSUPPORTED;
	
	if (!phid)
		return EPHIDGET_INVALIDARG;
	
	if (!CPhidget_statusFlagIsSet(phid->status, PHIDGET_ATTACHED_FLAG)
		&& !CPhidget_statusFlagIsSet(phid->status, PHIDGET_ATTACHING_FLAG))
		return EPHIDGET_NOTATTACHED;
	
	if (phid->deviceHandle == NULL)
	{
		LOG(PHIDGET_LOG_WARNING,"Handle for writing is not valid");
		return EPHIDGET_UNEXPECTED;
	}
	
	return CPhidgetAndroid_write(phid, buffer);
}

int CUSBSetLabel(CPhidgetHandle phid, char *buffer)
{
	int size = buffer[0];
	
	if(!ANDROID_USB_GOOD)
		return EPHIDGET_UNSUPPORTED;
	
	if(size>22) return EPHIDGET_INVALID;
	
	if (!phid)
		return EPHIDGET_INVALIDARG;
	
	if (!CPhidget_statusFlagIsSet(phid->status, PHIDGET_ATTACHED_FLAG)
		&& !CPhidget_statusFlagIsSet(phid->status, PHIDGET_ATTACHING_FLAG))
		return EPHIDGET_NOTATTACHED;
	
	if (phid->deviceHandle == NULL)
	{
		LOG(PHIDGET_LOG_WARNING,"Handle for writing is not valid");
		return EPHIDGET_UNEXPECTED;
	}

	return CPhidgetAndroid_setLabel(phid, buffer);
}

/* Buffer should be at least 8 bytes long */
int CUSBReadPacket(CPhidgetHandle phid, unsigned char *buffer) {
	if(!ANDROID_USB_GOOD)
		return EPHIDGET_UNSUPPORTED;
	
	if (!phid)
		return EPHIDGET_INVALIDARG;
	
	if (!CPhidget_statusFlagIsSet(phid->status, PHIDGET_ATTACHED_FLAG)
		&& !CPhidget_statusFlagIsSet(phid->status, PHIDGET_ATTACHING_FLAG))
		return EPHIDGET_NOTATTACHED;
	
	if (phid->deviceHandle == NULL)
	{
		LOG(PHIDGET_LOG_WARNING,"Handle for writing is not valid");
		return EPHIDGET_UNEXPECTED;
	}
	
	return CPhidgetAndroid_read(phid, buffer);
}

int CUSBRefreshLabelString(CPhidgetHandle phid)
{
	if(!ANDROID_USB_GOOD)
		return EPHIDGET_UNSUPPORTED;
	
	return CPhidgetAndroid_refreshLabel(phid);
}


int CUSBBuildList(CPhidgetList **curList)
{    
	if(!ANDROID_USB_GOOD)
		return EPHIDGET_UNSUPPORTED;
	
	return CPhidgetAndroidManager_getPhidgetList(curList);
}

void CUSBCleanup(void)
{
	;
}

int CUSBOpenHandle(CPhidgetHandle phid)
{  	
	if(!ANDROID_USB_GOOD)
		return EPHIDGET_UNSUPPORTED;
	
	return CPhidgetAndroid_openPhidget(phid);
}
